#include "genSrc/OrthBandpassFilter.h"
#include "genSrc/OrthBandpassFilter_scope.h"
#include <emC/Ipc/SocketComm_Ipc_emC.h>
#include <emC/Ipc/DatagramHead_Ipc_emC.h>
#include <stdio.h>
#include <stdlib.h>

// Hint: for this example a part of emC sources are used,
// especially for OSAL access, socket communication. 
// This is a common approach, should be run on any platform
// see https://vishia.org/emc/index.html

/**Data if the graphical given module, the type is code generated. */
OrthBandpassFilter_s filter;

/**Data for the test environment, this is manually written,
 * but the here used Scope_scope1_OrthBandpassFilter_s is generated. 
 */
struct MainData_T {
  float x;
  float xfq;

  /**This data are exchanged via ethernet, transmitted to the CurveViewApp. */
  struct ScopeData_T {
    DatagramHead_Ipc_emC_s head;    // it should be start on a even address dividable by 4.
    DatagramItem_Ipc_emC_s item1;
    int32BigEndian timesh;          // organization of time axis
    floatBigEndian time_ms;
    floatBigEndian x;               // the input value for test.
    Scope_scope1_OrthBandpassFilter_s scope1; // all other shown data from the graphic

  } scopeData;

  /**This data are exchanged via ethernet, received from the CurveViewApp. 
   * for a proper handshake of data exchange. */
  struct ScopeRxData_T {
    DatagramHead_Ipc_emC_s head;    // it should be start on a even address dividable by 4.
    DatagramItem_Ipc_emC_s item1;   // this contains sequence numbers for answer.
    int32BigEndian anyValue[3];     // not used but possible to use, part of payload in ethernet.
    int32BigEndian reserve[16];
  } scopeRxData;

} data = {0};

/**The Ethernet communication driver for PC. It is an OSAL-using part,
 * means can run under Windows as also under Linux, only by using the proper implementation files. */
SocketComm_Ipc_emC_s scope;

//helper
static int stop () { return 0; }

//Possibility to allocate data, experience.
//OrthBandpassFilter_s* pdata = (OrthBandpassFilter_s*)malloc(sizeof(OrthBandpass16_Ctrl_emC_s));

/**Main is also the test program. First only simple without specific test cases. */
int main(void){
  float Tstep = 0.000050f;
  ctor_OrthBandpassFilter(&filter, Tstep, 20.0f, 20.0f, 0.12f); // calls generated code
  //
  // The necessare and recommended init loop
  bool bInitOk = true;
  int ctAbortInit = 100;
  do {
    bool bInitOkPart;
    bInitOkPart = init_OrthBandpassFilter(&filter, 50.0f);      // calls generated code
    bInitOk &= bInitOkPart;
  } while(!bInitOk && --ctAbortInit >=0);
  if(ctAbortInit <0) {
    printf("ERROR init not successfull, abort\n");
    return 255;
  }
  ctor_SocketComm_Ipc_emC(&scope, 0x7f000001, 0xaef0 );  //127.0.0.1:44784 own address
  bool scopeOk = open_SocketComm_Ipc_emC(&scope, &data.scopeRxData, sizeof(data.scopeRxData), 50000);
  if(scopeOk) {  // set head data for the scope communication.
    setAddrDst_SocketComm_Ipc_emC(&scope, 0x7f000001, 0xaef1);      //127.0.0.1:44785 destination address
    setInt16BigEndian(&data.scopeData.head.nrofBytes, sizeof(data.scopeData));
    setInt16BigEndian(&data.scopeData.head.cmd, (int16)0xaeea);
    setInt16BigEndian(&data.scopeData.item1.nrofBytes, sizeof(data.scopeData)- sizeof(data.scopeData.head));
    setInt16BigEndian(&data.scopeData.item1.cmd, (int16)0xaeeb);
  }
  // also a file is possible to use for the curveViewAppl, but not yet used, can be used.
  FILE* fData = fopen("..\\..\\..\\src\\ExmplBandpassFilter\\data\\fq1.csv", "wb");
  if(fData !=null) { fprintf(fData, "timesh; time_ms; x; y1_re; y1_im; y3_re; y3_im; xd_re; xd_im; ym; yg; yph; yph2; fd; xfq; fq1");}
  data.xfq = 50.0f;
  //
  int nstepEnd = 20000; //runs for 1 second simulation time.
  int step = 0;
  float angle = 0;
  float angle3 = 0;
  float dangle3 = 3* PI_float_emC / 200.0f;  // 400 steps in 2*pi for fundamental wave
  //
  // This is the while loop for the simulation steps.
  while(++step < nstepEnd){
    upd_OrthBandpassFilter(&filter);   // calls generated code, update comes first.
    //
    float dangle = data.xfq * 2*PI_float_emC * Tstep;  // 400 steps in 2*pi for 50 Hz
    if(step == 7000){                  // disturbation:
      angle -= 0.1f;                   // angle step of ~5.8 ° because of load
    }
    data.x = sinf(angle += dangle) + 0.05f *cosf(angle3 += 3* dangle);
    if(step >= 12000 && step <=13500) {// frequency drift from 50.0 Hz to 49.8 Hz
      data.xfq -= 0.2f /500;           // 0.2 Hz in 500 steps
    }
    if(step >= 16900 && step <=20000) { // disturbation
      if(data.x > 0.9f) {               // it is similar a transformer saturation
        data.x += 0.5f * (data.x - 0.9f);
      }
    }
    if(angle >= PI_float_emC) { angle -= 2*PI_float_emC; }  // angle in range -pi..pi
    if(step == 9001){
      stop();                          // possibility to set a break point to view data
    }
    step_OrthBandpassFilter(&filter, data.x, 0);  // call of generated code
    //
    // ================================== with reduction ratio 1:32
    if( (step & 0x1f) ==1) {         
      updparam_OrthBandpassFilter(&filter);  // call param update
      param_OrthBandpassFilter(&filter, 50.0f);  // and call param - generated code
    }  
    if(scope.openOk) {  //=============== output to scope in any step time (fast)
//      if(hasRxTelg(&scope)) {   //...experience
//        int32 seqnr = getInt32BigEndian(&data.scopeRxData.seqnr);
//        setSeqnrRx(&scope, (int16)seqnr);        // inform the telegram synchronization about received seqnr.
//      }
      // ------------------------------------------ first counts back to clear Data
      setInt32BigEndian(&data.scopeData.head.seqnr, step);
      setInt32BigEndian(&data.scopeData.timesh, step<4 ? -step : step);
      setFloatBigEndian(&data.scopeData.time_ms, step*0.00005f);
      setFloatBigEndian(&data.scopeData.x, data.x);
      //--------------------------------- call of generated code for scope values.
      sendValues_scope1_OrthBandpassFilter(&data.scopeData.scope1, &filter);
      scopeOk = sendToDst_SocketComm_Ipc_emC (&scope, &data.scopeData, sizeof(data.scopeData), (int16)step, 100);
    }
    if(fData !=null) {
//    fprintf(fData,"\n%d; %1.5f; %1.4f; %1.4f; %1.4f; %1.4f; %1.4f; %1.4f; %1.4f; %1.4f; %1.4f; %1.4f; %1.4f; %1.4f; %1.4f; %1.4f;"
//      , step, step*0.00005f
//      , data.x
//      , filter.scope.yabz1.re    // has used older data struct, not adaped yet
//      , filter.scope.yabz1.im
//      , filter.scope.yabz3.re
//      , filter.scope.yabz3.im
//      , filter.scope.xdabz.re
//      , filter.scope.xdabz.im
//      , filter.scope.ym
//      , filter.scope.yg6
//      , filter.scope.yph3
//      , filter.scope.yph2
//      , filter.scope.fd
//      , data.xfq
//      , filter.scope.fq1
//      );
    }
  }
  if(fData !=null){
    fprintf(fData, "\n");
    fclose(fData);
  }
  close_SocketComm_Ipc_emC(&scope);
  printf("end\n");
  return 0;
}


#ifndef DEF_NO_THCXT_STACKTRC_EXC_emC
void uncatched_Exception_emC ( Exception_emC* ythis, ThreadContext_emC_s* _thCxt){
  printf("ERROR");
}
#endif

void errorSystem_emC_  (  int errorCode, const char* description, int value1, int value2, char const* file, int line){
  printf("ERROR");
}
