/************************************************************************************************
 * Copyright/Copyleft:
 *
 * For this source the LGPL Lesser General Public License,
 * published by the Free Software Foundation is valid.
 * It means:
 * 1) You can use this source without any restriction for any desired purpose.
 * 2) You can redistribute copies of this source to everybody.
 * 3) Every user of this source, also the user of redistribute copies
 *    with or without payment, must accept this license for further using.
 * 4) But the LPGL ist not appropriate for a whole software product,
 *    if this source is only a part of them. It means, the user
 *    must publish this part of source,
 *    but don't need to publish the whole source of the own product.
 * 5) You can study and modify (improve) this source
 *    for own using or for redistribution, but you have to license the
 *    modified sources likewise under this LGPL Lesser General Public License.
 *    You mustn't delete this Copyright/Copyleft inscription in this source file.
 *
 * This source may be used also with another licence, if the author
 * and all other here named co-authors have agreed to this contract.
 * Especially a company can use a copy of this sources in its products without publishing.
 * The user should have a underwritten contract therefore.
 *
 * @author This OS-Implementation was created by Alfred Schaffrina anno 2005/2006.
 *         It was used and improved by Hartmut Schorrig, www.vishia.org in 2007/2008.
 *         Some editing and improvements were done in 2008 by Marcos Rodriguez.
 *         The improvement after 2008 was done by Hartmut Schorrig.
 *
 **copyright***************************************************************************************
 *
 * @changes
 * 2024-03-15 Hartmut corr: Seems to be that it was not frequently used. Now fixed, used in Cygwin environment. Datagram socket runs.
 * 2010-01-31 Hartmut corr: use int..BigEndian in OS_SOCKADDR because it should be big endian.
 * 2010-01-15 Hartmut corr: �berarbeitung der Socket-Adaption f�r Hynet mit Auswirkungen hier, meist Feinheiten.
 *
 */


#include <emC/OSAL/os_socket.h>
#include <emC/Base/Endianness_emC.h>
//#include "os_error.h"
#include <sys/socket.h>
#include <errno.h>
#include <string.h>  //memset
//#include <linux/in.h> //Linux internet
#include <sys/socket.h>
#include <netdb.h>                 // see https://stackoverflow.com/questions/9102453/sockaddr-in-undeclared-identifier
#include <unistd.h>   //for close,
#include <fcntl.h>

//struct sockaddr_in;


// IPv6 AF_INET6 sockets:
#if 0
struct sockaddr_in6 {
    u_int16_t       sin6_family;   // address family, AF_INET6
    u_int16_t       sin6_port;     // port number, Network Byte Order
    u_int32_t       sin6_flowinfo; // IPv6 flow information
    struct XXXin6_addr sin6_addr;     // IPv6 address
    u_int32_t       sin6_scope_id; // Scope ID
};
#endif

/**This struct contains the implementation specific socket addr, it is compatible to OS_SOCKADDR, can be done a pointer cast.
 */
typedef struct Impl_SOCKADDR_T {
  int32 address1;  // same as in OS_SOCKADDR, first port 16 bit
  int32 address2;  // then in 4 byte

  /**Internal data defintion for linux adaption of sockets. It should have 16 bit, max. possible is 24 */
  union { struct sockaddr_in in; struct sockaddr addr; } linux;

  char dummy[24 - sizeof(struct sockaddr_in)];

} Impl_SOCKADDR;

#define bzero(ptr,len)  memset( (void *) ptr, (int) 0, (int) len )



#define negativeSign 0x80000000  //(1<<((sizeof(int)*8)-1))


/*
void copyAddrTo_SOCKADDR(struct sockaddr_in const* addrsys, OS_SOCKADDR* addr_OSAL)
{ //note:
  addr_OSAL->sin_port = (int16BigEndian)addrsys->sin_port;
  //A change of byte order may possible here. But it is both big endian.
  addr_OSAL->ip.add32 = (int32BigEndian)addrsys->sin_addr;
  //NOTE: Nucleus-net knows a char* name as 'machine's name', it isn't use here.
}


void copyAddrTo_addr_struct(OS_SOCKADDR const* addr_OSAL, struct sockaddr * addr_NU)
{ //note:
  addr_NU->family = NU_FAMILY_IP;
  addr_NU->port = addr_OSAL->sin_port;
  //NOTE: Nucleus-net knows constant MAX_ADDRESS_SIZE. Here only ip-4 is supported.
  //NOTE: In OS-Layer the first ip address should be located in that byte, which is the MSB in a normal int32 representation.
  //A change of byte order may possible here. But in the yet known implementatin it isn't regarded well.
  addr_NU->id.is_ip_addrs[0] = addr_OSAL->ip.sin_addr[0];
  addr_NU->id.is_ip_addrs[1] = addr_OSAL->ip.sin_addr[1];
  addr_NU->id.is_ip_addrs[2] = addr_OSAL->ip.sin_addr[2];
  addr_NU->id.is_ip_addrs[3] = addr_OSAL->ip.sin_addr[3];
}

*/


/**Initializes a OS_SOCKADDR from an Address_InterProcessComm. */
void set_OS_SOCKADDR  (  OS_SOCKADDR* dst, uint32 ip, uint16 port) {
  Impl_SOCKADDR* thiz = (Impl_SOCKADDR*) dst;
  dst->address1 = port;
  dst->address2 = ip;
  memset(dst->internalData, 0, sizeof(dst->internalData));
  //Note: cast to int16BigEndian* etc. is necessary, because the sin_family etc. are big endian but are not defined as such.
  thiz->linux.in.sin_family = AF_INET;
  //setInt16BigEndian((int16BigEndian*)&thiz->linux.in.sin_family, (int16)AF_INET);
  setInt16BigEndian((int16BigEndian*)&thiz->linux.in.sin_port, (int16) port);
  setInt32BigEndian((int32BigEndian*)&thiz->linux.in.sin_addr, ip);
}




int os_createStreamSocket(OS_StreamSocket* so )
{
  int ret;
  int hSocket;
  //WSAStartup is done, now create a socket:
  hSocket = socket(AF_INET, SOCK_STREAM, 0);
  if(hSocket == -1)
  { ret = errno | negativeSign; //mark with negativ sign! Keep the rest of hex-readable value.
    so->socket.socket = OS_INVALID_SOCKET;
  } else {
    ret = 0;
    so->socket.socket = (int32)hSocket; //it is a socket.
  }
  return ret;
}



int os_createDatagramSocket(OS_DatagramSocket* so)
{
  int ret;
  int hSocket;
  //WSAStartup is done, now create a socket:
  hSocket = socket(AF_INET, SOCK_DGRAM, 0);
  if(hSocket == -1)
  { ret = errno | negativeSign; //mark with negativ sign! Keep the rest of hex-readable value.
    so->socket.socket = OS_INVALID_SOCKET;
  } else {
    ret = 0;
    so->socket.socket = (int32)hSocket; //it is a socket.
  }
  return ret;
}


int os_createRawSocket  (  OS_RawSocket * so) {
  int ret;
  int hSocket;
  //WSAStartup is done, now create a socket:
  hSocket = socket(AF_INET, SOCK_RAW, 0);
  if(hSocket == -1)
  { ret = errno | negativeSign; //mark with negativ sign! Keep the rest of hex-readable value.
    so->socket.socket = OS_INVALID_SOCKET;
  } else {
    ret = 0;
    so->socket.socket = (int32)hSocket; //it is a socket.
  }
  return ret;
}



int os_bind(OS_Socket so, OS_SOCKADDR const* addr) {
  Impl_SOCKADDR const* addrL = (Impl_SOCKADDR const*) addr;
  int socklen = sizeof(*addr);

  int ok = bind(so.socket, &addrL->linux.addr, socklen);
  if(ok==-1){
      ok = errno | negativeSign;  //mark with negative sign! Keep the rest of hex-readable value.
  }
  return ok;
}


//instead os_bind, new version with or without rx timeout
int bind_OS_Socket  (  OS_Socket so, OS_SOCKADDR const* addr, int rxTimeout_us) {
  int error = os_bind(so, addr);
  if(error == 0 && rxTimeout_us >0) {   // see https://stackoverflow.com/questions/15941005/making-recvfrom-function-non-blocking
    struct timeval read_timeout;
    read_timeout.tv_sec = 0;
    read_timeout.tv_usec = rxTimeout_us;
    error = setsockopt(so.socket, SOL_SOCKET, SO_RCVTIMEO, (char const*)&read_timeout, sizeof read_timeout);
  }
  if(error == -1) {
    error |= negativeSign; //mark with negativ sign! Keep the rest of hex-readable value.
    so.socket = OS_INVALID_SOCKET;
  } else {
    error = 0;
  }
  return error;
}



int os_accept(OS_ServerSocket so, OS_SOCKADDR* dstAddr, OS_StreamSocket* soTcp) {
  Impl_SOCKADDR* addrL = (Impl_SOCKADDR*) dstAddr;
  int dstLen = sizeof(addrL->linux.in);
  int tcpSocket = accept(so.socket.socket, &addrL->linux.addr, &dstLen);
  if(tcpSocket !=-1){
    soTcp->socket.socket = tcpSocket;
    return 0;
  } else {
    soTcp->socket.socket = OS_INVALID_SOCKET;
    return errno | negativeSign;  //mark with negative sign! Keep the rest of hex-readable value.
  }
}




int os_close(OS_Socket so)
{
  int ok = close(so.socket);
  if(ok==-1){
      ok = errno | negativeSign;  //mark with negative sign! Keep the rest of hex-readable value.
  }
  return ok;
}




int os_connect(OS_StreamSocket so, OS_SOCKADDR const* dstAddr)
{
  Impl_SOCKADDR* addrL = (Impl_SOCKADDR*) dstAddr;
  int ok = connect( so.socket.socket, &addrL->linux.addr, sizeof(addrL->linux));
  if(ok==-1){
      ok = errno | negativeSign;  //mark with negative sign! Keep the rest of hex-readable value.
  }
  return ok;
}





int os_listen(OS_ServerSocket so, int backlog)
{
  int ok = listen( so.socket.socket, backlog);
  if(ok==-1){
      ok = errno | negativeSign;  //mark with negative sign! Keep the rest of hex-readable value.
  }
  return ok;
}







int os_recv(OS_StreamSocket so, void* buffer, int len, int flags)
{
  int ok = recv( so.socket.socket, (char*)buffer, len, flags);
  if(ok < 0){
      ok = errno | negativeSign;  //mark with negative sign! Keep the rest of hex-readable value.
  }
  return ok;

}





int os_recvfrom(OS_DatagramSocket so, void* buffer, int len, int flags, OS_SOCKADDR* from)
{
  Impl_SOCKADDR* addrL = (Impl_SOCKADDR*) from;
  int lenAddr = sizeof(addrL->linux.in);
  int ok = recvfrom( so.socket.socket, (char*)buffer, len, flags, &addrL->linux.addr, &lenAddr);
  if(ok < 0){
      ok = errno | negativeSign;  //mark with negative sign! Keep the rest of hex-readable value.
  }
  from->address1 = getInt16BigEndian((int16BigEndian*)&addrL->linux.in.sin_port);  //Note: casting necessary because sin_port is big endian but not declared as such.
  from->address2 = getInt32BigEndian((int32BigEndian*)&addrL->linux.in.sin_addr);
  return ok;
}





int os_send(OS_StreamSocket so, void const* buffer, int len, int flags)
{
  int ok = send( so.socket.socket, (const char*)buffer, len, flags);
  if(ok < 0){
      ok = errno | negativeSign;  //mark with negative sign! Keep the rest of hex-readable value.
  }
  return ok;
}





int os_sendto(OS_DatagramSocket so, void const* buffer, int len, int flags, OS_SOCKADDR const* to)
{
  Impl_SOCKADDR* addrL = (Impl_SOCKADDR*) to;

  size_t lenAddr = sizeof(addrL->linux);
  int ok = sendto( so.socket.socket, (const char*)buffer, len, flags, &addrL->linux.addr, lenAddr);
  if(ok < 0){
      ok = errno | negativeSign;  //mark with negative sign! Keep the rest of hex-readable value.
  }
  return ok;
}




/**Sets the just now opend socket to blocking or non blocking mode.
 * The implementation calls a special ioctrl routine from the system layer.
 * The standard behavior of a socket is the blocking mode. But this routine
 * should be called everytime after os_open to assure that the socket is in the expected mode.
 * @param af The Socket.
 * @param bNonBlocking If true than sets to non blocking, if false than sets to blocking.
 */
int os_setNonBlocking(OS_DatagramSocket socket, bool bNonBlocking)
{ int retVal;
  if(bNonBlocking)
  { retVal = fcntl(socket.socket.socket, O_NONBLOCK);
  }
  else
  { //blocking mode is the standard behavior ...
    retVal = 0;
  }
  return retVal;
}




int os_shutdown (OS_Socket so, int how)
{
        int ok = 0;

        switch(how){
        case SO_RECEIVE_DOWN:
                how = SHUT_RD;
                break;
        case SO_SEND_DOWN:
                how = SHUT_WR;
                break;
        case SO_BOTH_DOWN:
                how = SHUT_RDWR;
                break;
        default:
                break;
        }
        ok = shutdown( so.socket, how );
        if(ok < 0){
            ok = errno | negativeSign;  //mark with negative sign! Keep the rest of hex-readable value.
        }
        return ok;
}



//#include <asm-generic/errno.h>

const char* os_translateSocketErrorMsg(int nError)
{
  const char* sError;
  switch(nError & ~negativeSign)
  {
    case EPROTONOSUPPORT: sError = "Protocol not supported"; break;
    case 0x7fffffff: sError = "Permission denied."; break;
    default: sError = "unknown socket error";
  } //switch
  return sError;
}

